"""
Sprite
"""

import pygame, Numeric, pickle, math, copy
from pygame.locals import *

import sprite_repository
from sprites import my_rect

class my_sprite:
    def __init__(self, game, loc, sr_name, initial_facing, initial_state, sprite_collides, tile_collides):
        self.game = game
        self.sr_s = game.sr.sprites[sr_name]

        self.what_under = 0 # 0 -> Nothing, 1 -> Full obst. tile, 2 -> Part. obst. tile, 3 -> Obst. Sprite 
        self.sprite_under = None
        self.sprite_bump = False

        self.r = my_rect(loc[0], loc[1], self.sr_s.width, self.sr_s.height)
        self.facing = initial_facing # 0 is Right, 1 is Left
        self.vel = [0, 0]
        self.set_state(initial_state)
        
        self.sprite_collides = sprite_collides
        self.tile_collides = tile_collides

        self.cr_flags = dict()
        self.cr_flags['Vulnerable'] = False
        self.cr_flags['Attack'] = False

    def update(self):
        """Update the sprite, this can be extended by a derived sprite class (such as the Player) if necessary."""        
        # Save bounding rectangle
        self.old_r = copy.copy(self.r)
        # Handle sprite kinematics
        self.vel[1] = min(self.vel[1] + 0.125, 10) # Gravity
        if self.what_under == 3: # If standing on moving sprite           
            self.r.x += self.vel[0] + self.sprite_under.vel[0]
            self.r.y += min(self.vel[1] + self.sprite_under.vel[1], 10) # Terminal velocity!
        else:
            self.r.x, self.r.y = self.r.x + self.vel[0], self.r.y + self.vel[1]

        # Collision detection and response
        self.cr_collide()
        if self.tile_collides:
            self.map_collide()
        if self.sprite_collides:
            self.sprite_collide()

        # State update
        self.update_state()

        # Graphics update
        self.update_graphical()

    def update_state(self):
        """Each sprite class must implement a state machine here."""
        raise Exception("my_sprite.update_sprite() called!")

    def update_graphical(self):
        """Update the graphical aspect of the sprite."""       
        if not self.cur_anim or self.cur_anim.type == "None":
            return
        
        if pygame.time.get_ticks() > self.last_time + self.cur_anim.frames[self.cur_frame].time:
            self.last_time = pygame.time.get_ticks()
            self.cur_frame += self.frame_delta
            if self.cur_anim.type == "Ping-Pong":
                if self.cur_frame == 0 or self.cur_frame >= len(self.cur_anim.frames)-1:
                    self.frame_delta = -self.frame_delta
                    self.anim_runs += 1
            elif self.cur_anim.type == "Cyclic":
                if self.cur_frame >= len(self.cur_anim.frames):
                    self.cur_frame = 0
                    self.anim_runs += 1
            self.refresh_frame_vars()

    def set_state(self, new_state, frame=0):
        """Set the state of the sprite."""       
        self.state = new_state
        self.cur_anim = self.sr_s.animations[new_state]
        self.anim_runs = 0
        self.cur_frame = frame
        self.frame_delta = 1
        self.last_time = pygame.time.get_ticks()
        self.refresh_frame_vars()
        
    def get_cur_frame(self):
        """Shortcut to the current frame of the sprite"""
        if self.cur_anim:
            return self.cur_anim.frames[self.cur_frame]
        else:
            return None

    def draw(self, surf):
        """Draw the sprite to a surface."""      
        d = self.game.view_transform(self.r, self.o)
        if self.facing == 0:
            surf.blit(self.cur_anim.frames[self.cur_frame].img, d)
        else:
            surf.blit(self.cur_anim.frames[self.cur_frame].flip_img, d)

        # Debugraphics
        if False:
            v = self.game.view_transform(self.r)
#            pygame.draw.rect(surf, [31, 127, 127], v, 1)

            if self.facing == 0:
                c = self.get_cur_frame().collision_rects
            else:
                c = self.get_cur_frame().flip_collision_rects
            for i in c:
                if i.type == "Vulnerable":
                    colour = [0, 0, 127]
                elif i.type == "Attack":
                    colour = [127, 0, 0]
                pygame.draw.rect(surf, colour, (v[0] + i.rect.x, v[1] + i.rect.y, i.rect.w, i.rect.h), 1)

            if self.facing == 0:
                k = self.get_cur_frame().cr_bounds
            else:
                k = self.get_cur_frame().cr_flip_bounds
            pygame.draw.rect(surf, [127, 127, 127], (v[0] + k.x, v[1] + k.y, k.w, k.h), 1)
         
            surf.blit(self.game.game_font.render(`self.vel`, False, [63, 255, 255]), (v[0], v[1] - 12))
            surf.blit(self.game.game_font.render(self.state, False, [63, 255, 255]), (v[0], v[1] + self.r.h))
            surf.blit(self.game.game_font.render(`(self.what_under, self.last_time)`, False, [63, 255, 255]), (v[0], v[1] + self.r.h + 12))

    def refresh_frame_vars(self):
        """Refresh the attributes related to this frame of the sprite's animation."""
        self.o = (self.get_cur_frame().offset[0], self.get_cur_frame().offset[1])
        if self.facing == 1: # Flip horizontal component of offset if sprite is facing left
            visual_width = self.cur_anim.frames[self.cur_frame].img.get_width()
            self.o = (self.r.w - (visual_width + self.o[0]), self.o[1])

    def cr_collide(self):
        """Sets flags in this sprite according to the current state of overlap between it and the other sprites."""
        self.cr_flags['Vulnerable'] = False
        self.cr_flags['Attack'] = False
        
        for s in self.game.sl:
            # Don't collide with self!
            if s == self:
                continue

            # Fetch bounding rectangles
            if self.facing == 0:
                b = self.get_cur_frame().cr_bounds
                self_collision_rects = self.get_cur_frame().collision_rects
            else:
                b = self.get_cur_frame().cr_flip_bounds
                self_collision_rects = self.get_cur_frame().flip_collision_rects
            
            b1 = Rect(b)
            b1.x, b1.y = self.r.x + b1.x, self.r.y + b1.y
            b1 = self.game.view_transform(b1)

            b2 = Rect(b)
            b2.x, b2.y = s.r.x + b2.x, s.r.y + b2.y
            b2 = self.game.view_transform(b2)

            # Compare bounds
            if b1.colliderect(b2):
                # Get collision rects of s
                if s.facing == 0:
                    s_collision_rects = s.get_cur_frame().collision_rects
                else:
                    s_collision_rects = s.get_cur_frame().flip_collision_rects

                # Do test in O(N*M) time (could be reduced to O(N*log(m)) if need be)
                for self_cr in self_collision_rects:
                    for s_cr in s_collision_rects:
                        r1 = Rect(self_cr.rect.x, self_cr.rect.y, self_cr.rect.w, self_cr.rect.h)
                        r1.x, r1.y = self.r.x + r1.x, self.r.y + r1.y
                        r1 = self.game.view_transform(r1)

                        r2 = Rect(s_cr.rect.x, s_cr.rect.y, s_cr.rect.w, s_cr.rect.h)
                        r2.x, r2.y = s.r.x + r2.x, s.r.y + r2.y
                        r2 = self.game.view_transform(r2)
                        
                        if r1.colliderect(r2):
                            self.cr_flags[self_cr.type] = True
#                            s.cr_flags[s_cr.type] = True
        
    def map_collide(self):
        """Adjusts this sprite's rectangle so that it does not collide with the map."""
        # Exception if already colliding!
        if self.game.t.collides(self.old_r):
            raise Exception('Call to map_collide() with sprite already colliding! ' + str(self.old_r))

        # First handle collisions with fully obstructive tiles        
        if self.game.t.collides(self.r):
            o = [my_rect(self.old_r.x, self.r.y, self.r.w, self.r.h)]
            o.append(my_rect(self.r.x, self.old_r.y, self.r.w, self.r.h))
            o.append(my_rect(self.old_r.x, self.old_r.y, self.r.w, self.r.h))
            for a in range(len(o)):
                if not self.game.t.collides(o[a]):
                    if a == 0:
                        self.what_under = 0
                    else:
                        self.what_under = 1
                    self.r = o[a]
                    self.vel = [self.r.x - self.old_r.x, self.r.y - self.old_r.y]
                    break            

        # Now partial (bottom edge moving down and colliding with partial)
        a = my_rect(self.r.x, self.r.y + self.r.h, self.r.w, 0)
        if self.vel[1] > 0 and self.game.t.collides(a, True):
            a = my_rect(self.r.x, self.old_r.y + self.old_r.h, self.r.w, 0)
            if not self.game.t.collides(a, True):
                self.what_under = 2
                a = my_rect(self.r.x, self.old_r.y, self.r.w, self.old_r.h)
                self.r = a
                self.vel[1] = 0
                
    def sprite_collide(self):
        """Adjusts this sprite's rectangle so that it does not collide with the rectangle of any other sprite."""
        for s in self.game.sl:
            # Don't collide with self!
            if s == self:
                continue
            
            if s.r.collides(self.old_r) and (s.sprite_collides and self.sprite_collides):
                raise Exception('Call to sprite_collide() with sprite already colliding! ' + str(self.old_r))

            if s.r.collides(self.r):
                o = [my_rect(self.old_r.x, self.r.y, self.r.w, self.r.h)]
                o.append(my_rect(self.r.x, self.old_r.y, self.r.w, self.r.h))
                o.append(my_rect(self.old_r.x, self.old_r.y, self.r.w, self.r.h))
                for a in range(len(o)):
                    if not s.r.collides(o[a]):
                        if a == 0:
                            self.what_under = 0
                            self.sprite_under = None
                            self.sprite_bump = True
                            self.vel[0] = 0
                        elif a == 1:
                            self.what_under = 3
                            self.sprite_under = s
                            self.sprite_bump = False
                            self.vel[1] = 0
                        elif a == 2:
                            self.what_under = 3
                            self.sprite_under = None
                            self.sprite_bump = True
                            self.vel[0], self.vel[1] = 0, 0
                        self.r = o[a]
                        break
