"""
Tile Engine
"""

import pygame, Numeric, math
from pygame.locals import *

SCALE_2X = 1

class tile_engine:   
    def __init__(self, m, filename, flags=0):
        self.load_map(m)
        self.load_tileset(filename)

        if (flags & SCALE_2X):
            self.tileset_img = pygame.transform.scale(imag, (imag.get_width()*2, imag.get_height()*2))
            self.tile_size = 32
        else:
            self.tile_size = 16

        self.old_pos = None
        self.tileset_width = self.tileset_img.get_width() / self.tile_size
        self.tileset_height = self.tileset_img.get_height() / self.tile_size

    def load_map(self, m):
        self.map = m
        self.map_tile_width = len(self.map[0])
        self.map_tile_height = len(self.map)
        
    def load_tileset(self, filename):
        self.tileset = []
        self.tileset_img = pygame.image.load(filename + ".png")
        self.tileset_img = self.tileset_img.convert_alpha()

        import xml.dom.minidom
        d = xml.dom.minidom.parse(filename + ".xml")

        for t in d.documentElement.getElementsByTagName("tile"):
            cur_tile = tile(t.getAttribute("name"), t.getAttribute("obstruction"), t.getAttribute("animation"))            
            for f in t.getElementsByTagName("frame"):
                if cur_tile.animation == "None":
                    cur_tile.add_frame(int(f.getAttribute("index")))
                else:
                    cur_tile.add_frame(int(f.getAttribute("index")), int(f.getAttribute("time")))
            self.tileset.append(cur_tile)

    def update(self):
        for t in self.tileset:
            t.update()

    def slow_draw_map(self, surf, pos):
        for t_y in range(self.map_tile_height):
            for t_x in range(self.map_tile_width):
                i = self.tileset[self.map[t_y][t_x]].cur_index()
                c_x = (i % self.tileset_width) * self.tile_size
                c_y = (i / self.tileset_width) * self.tile_size
                t = pygame.Rect((c_x, c_y, self.tile_size, self.tile_size))
                surf.blit(self.tileset_img, (-pos[0] + t_x * self.tile_size, -pos[1] + t_y * self.tile_size), t)

    def draw_map(self, surf, pos, full):
        """
        o_O_o
        """       
        surf_tile_width = math.ceil(float(surf.get_width()) / float(self.tile_size)) + 1
        surf_tile_height = math.ceil(float(surf.get_height()) / float(self.tile_size)) + 1
        start_x = max([pos[0] // self.tile_size, 0])
        start_y = max([pos[1] // self.tile_size, 0])
        end_x = min([start_x + surf_tile_width, self.map_tile_width])
        end_y = min([start_y + surf_tile_height, self.map_tile_height])

        if self.old_pos != pos:
            full = 1 # slow temp hack, could redraw edge tiles only

        self.draw_region(surf, Rect(start_x, start_y, end_x-start_x, end_y-start_y), pos, full) # more slowness
        self.old_pos = pos

    def draw_region(self, surf, region, pos, full):
        for t_y in range(region.y, region.y + region.h):
            for t_x in range(region.x, region.x + region.w):
                if self.map[t_y][t_x] != -1:
                    t = self.tileset[self.map[t_y][t_x]]
                    if full or t.redraw:
                        l = (-pos[0] + t_x * self.tile_size, -pos[1] + t_y * self.tile_size)
                        self.draw_tile(surf, self.map[t_y][t_x], l)
                elif full:
                    l = (-pos[0] + t_x * self.tile_size, -pos[1] + t_y * self.tile_size)
#                    surf.fill(0, Rect(l[0], l[1], self.tile_size, self.tile_size))

    def draw_tile(self, surf, tile, loc):
        i = self.tileset[tile].cur_index()
        c_x = (i % self.tileset_width) * self.tile_size
        c_y = (i / self.tileset_width) * self.tile_size
        r = pygame.Rect((c_x, c_y, self.tile_size, self.tile_size))
#        surf.fill(0, pygame.Rect(loc[0], loc[1], self.tile_size, self.tile_size))
        surf.blit(self.tileset_img, loc, r)

    def make_minimap(self):
        self.minimap = pygame.Surface((self.map_tile_width, self.map_tile_height), HWSURFACE)
        self.minimap.fill((0, 0, 63))
        self.minimap.lock()

        for t_y in range(0, self.map_tile_height):
            for t_x in range(0, self.map_tile_width):
                if self.map[t_y][t_x] != -1:
                    if self.tileset[self.map[t_y][t_x]].obstruction == "Full":
                        self.minimap.set_at((t_x, t_y), (191, 191, 191))
                    elif self.tileset[self.map[t_y][t_x]].obstruction == "Partial":
                        self.minimap.set_at((t_x, t_y), (127, 127, 127))

        self.minimap.unlock()

    def collides(self, r, partial=False):
        """Returns whether a rectangle is colliding with the map or not."""
        tile_left = int(r.x) / self.tile_size
        tile_right = int(math.ceil(float(r.x + r.w) / self.tile_size)) - 1
        tile_bottom = int(math.ceil(float(r.y + r.h) / self.tile_size)) - 1
        tile_top = int(r.y) / self.tile_size

        for t_y in range(tile_top, tile_bottom + 1):
            for t_x in range(tile_left, tile_right + 1):
                i = self.map[t_y][t_x]
                t = self.tileset[i]

                if i != -1 and (t.obstruction == "Full" or (partial and t.obstruction == "Partial")):
                    return True
        return False        
        
    def cast_ray(self, p, d):
        """Casts a ray in the map.  Returns a pair containing the intersection point and it's type."""
        # Test for axis-aligned direction
        x_test = 1
        y_test = 1
        
        if d[0] == 0:
            y_test = 0
        if d[1] == 0:
            x_test = 0

        # Zero direction?
        if x_test == 0 and y_test == 0:
            return (p, 2)

        # Initial casts
        if y_test:
            y_slope = float(d[1]) / float(d[0]) # dy/dx
            y_delta = y_slope * self.tile_size

            if d[0] < 0:
                temp_x = p[0] / self.tile_size * self.tile_size
                x_dir = -1
            else:
                temp_x = int(math.ceil(float(p[0]) / float(self.tile_size)) * float(self.tile_size))
                x_dir = 1
            y_i = (temp_x, int(p[1] - (p[0] - temp_x) * y_slope))

        if x_test:
            x_slope = float(d[0]) / float(d[1]) # dx/dy
            x_delta = x_slope * self.tile_size

            if d[1] < 0:
                temp_y = p[1] / self.tile_size * self.tile_size
                y_dir = -1
            else:
                temp_y = int(math.ceil(float(p[1]) / float(self.tile_size)) * float(self.tile_size))
                y_dir = 1
            x_i = (int(p[0] - (p[1] - temp_y) * x_slope), temp_y)

        # Keep going until intersection found
        ### IMPLEMENT ME NOW! ###

        # Calculate distances
        if x_test:
            d_x = self.dist(p, x_i)
        if y_test:
            d_y = self.dist(p, y_i)

        # Return appropriate intersection
        if x_test and y_test and d_x == d_y:
            return (x_i, 2)      
        if not y_test or d_x < d_y:
            return (x_i, 0)
        elif not x_test or d_y < d_x:
            return (y_i, 1)        
            
        # Exception, no appropriate intersection exists
        return None
 
    def dist(self, a, b):
        d = (b[0] - a[0], b[1] - a[1])
        return math.sqrt(d[0]*d[0] + d[1]*d[1])
            
class tile:
    def __init__(self, name, obstr, anim):
        self.name = name
        self.obstruction = obstr
        self.animation = anim

        self.frames = []
        self.cur_frame = 0

        self.frame_delta = 1
        self.last_time = pygame.time.get_ticks()
        self.redraw = 1

    def update(self):        
        self.redraw = 0
        if self.animation == "None":
            return
        if pygame.time.get_ticks() > self.last_time + self.frames[self.cur_frame][1]:
            self.redraw = 1
            self.last_time = pygame.time.get_ticks()
            self.cur_frame += self.frame_delta
            if self.animation == "Ping-Pong":
                if self.cur_frame == 0 or self.cur_frame >= len(self.frames)-1:
                    self.frame_delta = -self.frame_delta
            elif self.animation == "Cyclic":
                if self.cur_frame >= len(self.frames):
                    self.cur_frame = 0

    def add_frame(self, index, time = 0):
        self.frames.append((index, time))

    def cur_index(self):
        return self.frames[self.cur_frame][0]        
